"""Run fully configured classification blueprint.

Customization via class attributes is possible. Configs can be instantiated and provided to change Xgboost training.
Default hyperparameter search space is relatively light-weight to speed up the prototyping.
Can deal with binary and multi-class classification problems.
Hyperparameter tuning can be switched off or even strengthened via cross-validation. This behaviour can be controlled
via the config class attributes from config.training_config module.
"""
import warnings
from datetime import datetime
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import numpy as np
import pandas as pd

from bluecast.config.training_config import (
    TrainingConfig,
    XgboostFinalParamConfig,
    XgboostTuneParamsConfig,
)
from bluecast.evaluation.eval_metrics import eval_classifier
from bluecast.evaluation.shap_values import shap_explanations
from bluecast.experimentation.tracking import ExperimentTracker
from bluecast.general_utils.general_utils import check_gpu_support, logger
from bluecast.ml_modelling.xgboost import XgboostModel
from bluecast.preprocessing.custom import CustomPreprocessing
from bluecast.preprocessing.datetime_features import date_converter
from bluecast.preprocessing.encode_target_labels import TargetLabelEncoder
from bluecast.preprocessing.feature_selection import RFECVSelector
from bluecast.preprocessing.feature_types import FeatureTypeDetector
from bluecast.preprocessing.nulls_and_infs import fill_infinite_values
from bluecast.preprocessing.schema_checks import SchemaDetector
from bluecast.preprocessing.target_encoding import (
    BinaryClassTargetEncoder,
    MultiClassTargetEncoder,
)
from bluecast.preprocessing.train_test_split import train_test_split


class BlueCast:
    """Run fully configured classification blueprint.

    Customization via class attributes is possible. Configs can be instantiated and provided to change Xgboost training.
    Default hyperparameter search space is relatively light-weight to speed up the prototyping.
    :param :class_problem: Takes a string containing the class problem type. Either "binary" or "multiclass".
    :param :target_column: Takes a string containing the name of the target column.
    :param :cat_columns: Takes a list of strings containing the names of the categorical columns. If not provided,
        BlueCast will infer these automatically.
    :param :date_columns: Takes a list of strings containing the names of the date columns. If not provided,
        BlueCast will infer these automatically.
    :param :time_split_column: Takes a string containing the name of the time split column. If not provided,
        BlueCast will not split the data by time or order, but do a random split instead.
    :param :ml_model: Takes an instance of a XgboostModel class. If not provided, BlueCast will instantiate one.
        This is an API to pass any model class. Inherit the baseclass from ml_modelling.base_model.BaseModel.
    :param custom_in_fold_preprocessor: Takes an instance of a CustomPreprocessing class. Allows users to eeecute
        preprocessing after the train test split within cv folds. This will be executed only if precise_cv_tuning in
        the conf_Training is True. Custom ML models need to implement this themselves. This step is only useful when
        the proprocessing step has a high chance of overfitting otherwise (i.e: oversampling techniques).
    :param custom_preprocessor: Takes an instance of a CustomPreprocessing class. Allows users to inject custom
        preprocessing steps which take place right after the train test spit.
    :param custom_last_mile_computation: Takes an instance of a CustomPreprocessing class. Allows users to inject custom
        preprocessing steps which take place right before the model training.
    :param experiment_tracker: Takes an instance of an ExperimentTracker class. If not provided this will be initialized
        automatically.
    """

    def __init__(
        self,
        class_problem: Literal["binary", "multiclass"],
        target_column: Union[str, float, int],
        cat_columns: Optional[List[Union[str, float, int]]] = None,
        date_columns: Optional[List[Union[str, float, int]]] = None,
        time_split_column: Optional[str] = None,
        ml_model: Optional[Union[XgboostModel, Any]] = None,
        custom_in_fold_preprocessor: Optional[CustomPreprocessing] = None,
        custom_last_mile_computation: Optional[CustomPreprocessing] = None,
        custom_preprocessor: Optional[CustomPreprocessing] = None,
        custom_feature_selector: Optional[
            Union[RFECVSelector, CustomPreprocessing]
        ] = None,
        conf_training: Optional[TrainingConfig] = None,
        conf_xgboost: Optional[XgboostTuneParamsConfig] = None,
        conf_params_xgboost: Optional[XgboostFinalParamConfig] = None,
        experiment_tracker: Optional[ExperimentTracker] = None,
    ):
        self.class_problem = class_problem
        self.prediction_mode: bool = False
        self.cat_columns = cat_columns
        self.date_columns = date_columns
        self.time_split_column = time_split_column
        self.target_column = target_column
        self.conf_training = conf_training
        self.conf_xgboost = conf_xgboost
        self.conf_params_xgboost = conf_params_xgboost
        self.feat_type_detector: Optional[FeatureTypeDetector] = None
        self.cat_encoder: Optional[
            Union[BinaryClassTargetEncoder, MultiClassTargetEncoder]
        ] = None
        self.target_label_encoder: Optional[TargetLabelEncoder] = None
        self.schema_detector: Optional[SchemaDetector] = None
        self.ml_model: Optional[XgboostModel] = ml_model
        self.custom_in_fold_preprocessor = custom_in_fold_preprocessor
        self.custom_last_mile_computation = custom_last_mile_computation
        self.custom_preprocessor = custom_preprocessor
        self.custom_feature_selector = custom_feature_selector
        self.shap_values: Optional[np.ndarray] = None
        self.eval_metrics: Optional[Dict[str, Any]] = None

        if experiment_tracker:
            self.experiment_tracker = experiment_tracker
        else:
            self.experiment_tracker = ExperimentTracker()

        if not self.conf_params_xgboost:
            self.conf_params_xgboost = XgboostFinalParamConfig()

    def initial_checks(self, df: pd.DataFrame) -> None:
        if not self.conf_training:
            self.conf_training = TrainingConfig()
        if not self.conf_training.enable_feature_selection:
            message = """Feature selection is disabled. Update the TrainingConfig param 'enable_feature_selection'
            to enable it or make use of a custom preprocessor to do it manually during the last mile computations step.
            Feature selection is recommended for datasets with many features (>1000). For datasets with a small amount
            of features feature selection is not recommended.
            """
            warnings.warn(message, UserWarning, stacklevel=2)

        if self.conf_training.hypertuning_cv_folds == 1:
            message = """Cross validation is disabled. Update the TrainingConfig param 'hypertuning_cv_folds'
            to enable it. Cross validation is disabled on default to allow fast prototyping. For robust hyperparameter
            tuning using at least 5 folds is recommended."""
            warnings.warn(message, UserWarning, stacklevel=2)

        if (
            self.conf_training.enable_feature_selection
            and not self.custom_feature_selector
        ):
            message = """Feature selection is enabled but no feature selector has been provided. Falling back to
            cross-validated feature elimination. Specifically for small datasets check the logs to verify that not too
            many features have been removed. Otherwise, consider disabling feature selection or providing a custom
            feature selector."""
            warnings.warn(message, UserWarning, stacklevel=2)
        if not self.conf_xgboost:
            message = """No XgboostTuneParamsConfig has been provided. Falling back to default values. Default values
            have been chosen to speed up the prototyping. For robust hyperparameter tuning consider providing a custom
            XgboostTuneParamsConfig with a deeper hyperparameter search space and a custom TrainingConfig to enable
            cross-validation."""
            warnings.warn(message, UserWarning, stacklevel=2)
        if (
            self.conf_training.min_features_to_select >= len(df.columns)
            and self.conf_training.enable_feature_selection
        ):
            message = """The minimum number of features to select is greater or equal to the number of features in
            the dataset while feature selection is enabled. Consider reducing the minimum number of features to
            select or disabling feature selection via TrainingConfig."""
            warnings.warn(message, UserWarning, stacklevel=2)
        if (
            self.conf_training.cat_encoding_via_ml_algorithm
            and self.conf_training.calculate_shap_values
        ):
            self.conf_training.calculate_shap_values = False
            message = """SHAP values cannot be calculated when categorical encoding via ML algorithm is enabled due to
            incompatibility with the shap library. See this GitHub issue for more context:
            https://github.com/slundberg/shap/issues/266

            Calculation of Shap values has been changed to false.
            Consider disabling categorical encoding via ML algorithm in the TrainingConfig if shap values are
            required. Alternatively use Xgboost as a custom model and calculate shap values manually via
            pred_contribs=True."""
            warnings.warn(message, UserWarning, stacklevel=2)
        if self.conf_training.cat_encoding_via_ml_algorithm and self.ml_model:
            message = """Categorical encoding via ML algorithm is enabled. Make sure to handle categorical features
            within the provided ml model or consider disabling categorical encoding via ML algorithm in the
            TrainingConfig alternatively."""
            warnings.warn(message, UserWarning, stacklevel=2)
        if (
            self.conf_training.cat_encoding_via_ml_algorithm
            and self.custom_last_mile_computation
        ):
            message = """Categorical encoding via ML algorithm is enabled. Make sure to handle categorical features
            within the provided last mile computation or consider disabling categorical encoding via ML algorithm in the
            TrainingConfig alternatively."""
            warnings.warn(message, UserWarning, stacklevel=2)
        if (
            self.conf_training.precise_cv_tuning
            and not self.custom_in_fold_preprocessor
        ):
            message = """Precise fine tuning has been enabled, but no custom_in_fold_preprocessor has been provided.
            This will cause long runtimes without benefit. If you plan to execute any overfitting risky preprocessing,
            please consider using custom_in_fold_preprocessor to execute the steps within the cross-validation folds
            using precise_cv_tuning. Otherwise disable precise_cv_tuning to benefit from early pruning of unpromising
            hyperparameter sets."""
            warnings.warn(message, UserWarning, stacklevel=2)
        if (
            self.conf_training.precise_cv_tuning
            and self.conf_training.hypertuning_cv_folds < 2
        ):
            message = """Precise fine tuning has been enabled, but number of hypertuning_cv_folds is less than 2. With
            less than 2 folds precise_cv_tuning will not have any impact. Consider raising the number of folds to two
            or higher or disable precise_cv_tuning."""
            warnings.warn(message, UserWarning, stacklevel=2)

    def fit(self, df: pd.DataFrame, target_col: str) -> None:
        """Train a full ML pipeline."""
        check_gpu_support()
        feat_type_detector = FeatureTypeDetector()
        df = feat_type_detector.fit_transform_feature_types(df)
        self.feat_type_detector = feat_type_detector

        if self.feat_type_detector.cat_columns:
            if self.target_column in self.feat_type_detector.cat_columns:
                self.target_label_encoder = TargetLabelEncoder()
                df.loc[
                    :, self.target_column
                ] = self.target_label_encoder.fit_transform_target_labels(
                    df.loc[:, self.target_column]
                )

        self.cat_columns = self.feat_type_detector.cat_columns
        self.date_columns = self.feat_type_detector.date_columns

        if not self.conf_training:
            self.conf_training = TrainingConfig()

        self.initial_checks(df)

        x_train, x_test, y_train, y_test = train_test_split(
            df,
            target_col,
            self.time_split_column,
            self.conf_training.train_size,
            self.conf_training.global_random_state,
            self.conf_training.train_split_stratify,
        )

        if self.custom_preprocessor:
            x_train, y_train = self.custom_preprocessor.fit_transform(x_train, y_train)
            x_test, y_test = self.custom_preprocessor.transform(
                x_test, y_test, predicton_mode=False
            )
            feat_type_detector = FeatureTypeDetector()
            _ = feat_type_detector.fit_transform_feature_types(x_train)
            x_train, y_train = x_train.reset_index(drop=True), y_train.reset_index(
                drop=True
            )
            x_test, y_test = x_test.reset_index(drop=True), y_test.reset_index(
                drop=True
            )
            if target_col in feat_type_detector.cat_columns:
                feat_type_detector.cat_columns.remove(target_col)

        x_train, x_test = fill_infinite_values(x_train), fill_infinite_values(x_test)
        x_train, x_test = date_converter(x_train, self.date_columns), date_converter(
            x_test, self.date_columns
        )

        self.schema_detector = SchemaDetector()
        self.schema_detector.fit(x_train)
        x_train = self.schema_detector.transform(x_train)
        x_test = self.schema_detector.transform(x_test)

        if (
            self.cat_columns is not None
            and self.class_problem == "binary"
            and not self.conf_training.cat_encoding_via_ml_algorithm
        ):
            self.cat_encoder = BinaryClassTargetEncoder(feat_type_detector.cat_columns)
            x_train = self.cat_encoder.fit_target_encode_binary_class(x_train, y_train)
            x_test = self.cat_encoder.transform_target_encode_binary_class(x_test)
        elif (
            self.cat_columns is not None
            and self.class_problem == "multiclass"
            and not self.conf_training.cat_encoding_via_ml_algorithm
        ):
            self.cat_encoder = MultiClassTargetEncoder(
                feat_type_detector.cat_columns, self.target_column
            )
            x_train = self.cat_encoder.fit_target_encode_multiclass(x_train, y_train)
            x_test = self.cat_encoder.transform_target_encode_multiclass(x_test)
        elif self.conf_training.cat_encoding_via_ml_algorithm:
            x_train[self.cat_columns] = x_train[self.cat_columns].astype("category")
            x_test[self.cat_columns] = x_test[self.cat_columns].astype("category")

        if self.custom_last_mile_computation:
            x_train, y_train = self.custom_last_mile_computation.fit_transform(
                x_train, y_train
            )
            x_test, y_test = self.custom_last_mile_computation.transform(
                x_test, y_test, predicton_mode=False
            )

        if not self.custom_feature_selector:
            self.custom_feature_selector = RFECVSelector(
                random_state=self.conf_training.global_random_state,
                min_features_to_select=self.conf_training.min_features_to_select,
            )

        if self.conf_training.enable_feature_selection:
            x_train, y_train = self.custom_feature_selector.fit_transform(
                x_train, y_train
            )
            x_test, _ = self.custom_feature_selector.transform(
                x_test, predicton_mode=False
            )

        if not self.ml_model:
            self.ml_model = XgboostModel(
                self.class_problem,
                conf_training=self.conf_training,
                conf_xgboost=self.conf_xgboost,
                conf_params_xgboost=self.conf_params_xgboost,
                experiment_tracker=self.experiment_tracker,
            )
        self.ml_model.fit(x_train, x_test, y_train, y_test)
        if self.conf_training and self.conf_training.calculate_shap_values:
            self.shap_values = shap_explanations(self.ml_model.model, x_test)
        self.prediction_mode = True

    def fit_eval(
        self,
        df: pd.DataFrame,
        df_eval: pd.DataFrame,
        target_eval: pd.Series,
        target_col: str,
    ) -> Dict[str, Any]:
        """Train a full ML pipeline and evaluate on a holdout set.

        This is a convenience function to train and evaluate on a holdout set. It is recommended to use this for model
        exploration. On production the simple fit() function should be used.
        :param :df: Takes a pandas DataFrame containing the training data and the targets.
        :param :df_eval: Takes a pandas DataFrame containing the evaluation data, but not the targets.
        :param :target_eval: Takes a pandas Series containing the evaluation targets.
        :param :target_col: Takes a string containing the name of the target column inside the training data df.
        """
        self.fit(df, target_col)
        y_probs, y_classes = self.predict(df_eval)

        if self.feat_type_detector:
            if self.target_label_encoder and self.feat_type_detector:
                eval_df = pd.DataFrame(target_eval.values, columns=[target_col])
                y_true = self.target_label_encoder.transform_target_labels(
                    eval_df, target_col
                )
            else:
                y_true = target_eval.values
        else:
            y_true = target_eval.values

        eval_dict = eval_classifier(
            y_true, y_probs, y_classes
        )  # TODO: add experiment tracker
        self.eval_metrics = eval_dict

        if not self.conf_training:
            raise ValueError("Could not find any training config")

        if not self.conf_params_xgboost:
            raise ValueError("Could not find Xgboost params")

        if len(self.experiment_tracker.experiment_id) == 0:
            self.experiment_tracker.experiment_id.append(0)

        # enrich experiment tracker
        for metric, higher_is_better in zip(
            [
                "accuracy",
                "recall",
                "f1_score_weighted",
                "log_loss",
                "balanced_logloss",
                "roc_auc",
                "matthews",
            ],
            [True, True, True, False, False, True, True],
        ):
            self.experiment_tracker.add_results(
                experiment_id=self.experiment_tracker.experiment_id[-1],
                score_category="oof_score",
                training_config=self.conf_training,
                model_parameters=self.conf_params_xgboost.params,  # noqa
                eval_scores=self.eval_metrics["accuracy"],
                metric_used=metric,
                metric_higher_is_better=higher_is_better,
            )
        return eval_dict

    def transform_new_data(self, df: pd.DataFrame) -> pd.DataFrame:
        """Transform new data according to preprocessing pipeline."""
        check_gpu_support()
        if not self.feat_type_detector:
            raise Exception("Feature type converter could not be found.")

        if not self.conf_training:
            raise Exception("Training configuration could not be found.")

        df = self.feat_type_detector.transform_feature_types(
            df, ignore_cols=[self.target_column]
        )

        if self.custom_preprocessor:
            df, _ = self.custom_preprocessor.transform(df, predicton_mode=True)
            df = df.reset_index(drop=True)

        df = fill_infinite_values(df)
        df = date_converter(df, self.date_columns)

        if self.schema_detector:
            df = self.schema_detector.transform(df)

        if (
            self.cat_columns
            and self.cat_encoder
            and self.class_problem == "binary"
            and isinstance(self.cat_encoder, BinaryClassTargetEncoder)
            and not self.conf_training.cat_encoding_via_ml_algorithm
        ):
            df = self.cat_encoder.transform_target_encode_binary_class(df)
        elif (
            self.cat_columns
            and self.cat_encoder
            and self.class_problem == "multiclass"
            and isinstance(self.cat_encoder, MultiClassTargetEncoder)
            and not self.conf_training.cat_encoding_via_ml_algorithm
        ):
            df = self.cat_encoder.transform_target_encode_multiclass(df)
        elif self.conf_training.cat_encoding_via_ml_algorithm:
            df[self.cat_columns] = df[self.cat_columns].astype("category")

        if self.custom_last_mile_computation:
            df, _ = self.custom_last_mile_computation.transform(df, predicton_mode=True)

        if self.custom_feature_selector and self.conf_training.enable_feature_selection:
            df, _ = self.custom_feature_selector.transform(df, predicton_mode=True)

        return df

    def predict(self, df: pd.DataFrame) -> Tuple[np.ndarray, np.ndarray]:
        """Predict on unseen data.

        Return the predicted probabilities and the predicted classes:
        y_probs, y_classes = predict(df)
        """
        if not self.ml_model:
            raise Exception("Ml model could not be found")

        if not self.feat_type_detector:
            raise Exception("Feature type converter could not be found.")

        if not self.conf_training:
            raise ValueError("conf_training is None")

        check_gpu_support()
        df = self.transform_new_data(df)

        logger(f"{datetime.utcnow()}: Predicting...")
        y_probs, y_classes = self.ml_model.predict(df)

        if self.feat_type_detector.cat_columns:
            if (
                self.target_column in self.feat_type_detector.cat_columns
                and self.target_label_encoder
                and self.feat_type_detector
            ):
                y_classes = self.target_label_encoder.label_encoder_reverse_transform(
                    pd.Series(y_classes)
                )

        return y_probs, y_classes



import pytest

from bluecast.config.training_config import TrainingConfig
from bluecast.experimentation.tracking import ExperimentTracker


@pytest.fixture
def experiment_tracker():
    return ExperimentTracker()


def test_add_results(experiment_tracker):
    # Add some sample data
    experiment_id = 1
    score_category = "cv_score"
    training_config = (
        TrainingConfig()
    )  # You may need to create a valid TrainingConfig instance
    model_parameters = {"param1": 1, "param2": "abc"}
    eval_scores = 0.95
    metric_used = "accuracy"
    metric_higher_is_better = True

    experiment_tracker.add_results(
        experiment_id,
        score_category,
        training_config,
        model_parameters,
        eval_scores,
        metric_used,
        metric_higher_is_better,
    )

    # Check if the data was added correctly
    assert experiment_tracker.experiment_id == [1]
    assert experiment_tracker.score_category == ["cv_score"]
    assert experiment_tracker.training_configs == [
        training_config.model_dump(mode="json")
    ]
    assert experiment_tracker.model_parameters == [model_parameters]
    assert experiment_tracker.eval_scores == [0.95]
    assert experiment_tracker.metric_used == ["accuracy"]
    assert experiment_tracker.metric_higher_is_better == [True]


def test_retrieve_results_as_df(experiment_tracker):
    # Add some sample data
    experiment_id = 1
    score_category = "cv_score"
    training_config = (
        TrainingConfig()
    )  # You may need to create a valid TrainingConfig instance
    model_parameters = {"param1": 1, "param2": "abc"}
    eval_scores = 0.95
    metric_used = "accuracy"
    metric_higher_is_better = True

    experiment_tracker.add_results(
        experiment_id,
        score_category,
        training_config,
        model_parameters,
        eval_scores,
        metric_used,
        metric_higher_is_better,
    )

    # Retrieve the results as a DataFrame
    results_df = experiment_tracker.retrieve_results_as_df()

    # Check if the retrieved data matches the added data
    assert results_df["experiment_id"].tolist() == [1]
    assert results_df["score_category"].tolist() == ["cv_score"]
    assert results_df["eval_scores"].tolist() == [0.95]
    assert results_df["metric_used"].tolist() == ["accuracy"]
    assert results_df["metric_higher_is_better"].tolist() == [True]


def test_get_best_score_empty(experiment_tracker):
    # Ensure it raises an exception when no results have been added
    with pytest.raises(
        ValueError, match="No results have been found in experiment tracker"
    ):
        experiment_tracker.get_best_score(target_metric="loss")


def test_get_best_score_higher_is_better(experiment_tracker):
    # Add some sample data with a higher-is-better metric
    experiment_id = 1
    score_category = "cv_score"
    training_config = (
        TrainingConfig()
    )  # You may need to create a valid TrainingConfig instance
    model_parameters = {"param1": 1, "param2": "abc"}
    eval_scores = 0.95
    metric_used = "accuracy"
    metric_higher_is_better = True

    experiment_tracker.add_results(
        experiment_id,
        score_category,
        training_config,
        model_parameters,
        eval_scores,
        metric_used,
        metric_higher_is_better,
    )

    # Ensure the best score is correctly computed
    best_score = experiment_tracker.get_best_score(target_metric="accuracy")
    assert best_score == 0.95


def test_get_best_score_lower_is_better(experiment_tracker):
    # Add some sample data with a lower-is-better metric
    experiment_id = 1
    score_category = "cv_score"
    training_config = (
        TrainingConfig()
    )  # You may need to create a valid TrainingConfig instance
    model_parameters = {"param1": 1, "param2": "abc"}
    eval_scores = 0.95
    metric_used = "loss"  # Assuming "loss" is a metric that is lower-is-better
    metric_higher_is_better = False

    experiment_tracker.add_results(
        experiment_id,
        score_category,
        training_config,
        model_parameters,
        eval_scores,
        metric_used,
        metric_higher_is_better,
    )

    # Ensure the best score is correctly computed
    best_score = experiment_tracker.get_best_score(target_metric="loss")
    assert best_score == 0.95



import warnings
from typing import List, Union

import pandas as pd

from bluecast.eda.analyse import theil_u


def detect_leakage_via_correlation(
    data: pd.DataFrame, target_column: Union[str, float, int], threshold: float = 0.9
) -> List[Union[str, float, int, None]]:
    """
    Detect data leakage by checking for high correlations between the target column
    and other columns in the DataFrame. The target column must be part of the provided
    DataFrame.

    :param data: The DataFrame containing the data (numerical columns only for features)
    :param target_column: The name of the target column to check for correlations.
    :param threshold: The correlation threshold. If the absolute correlation value is greater than
      or equal to this threshold, it will be considered as a potential data leakage.
    :returns: True if data leakage is detected, False if not.
    """
    if target_column not in data.columns:
        raise ValueError(
            f"The target column '{target_column}' is not found in the DataFrame."
        )

    correlations = data.corr()[target_column].abs()
    potential_leakage = correlations[correlations >= threshold].index.tolist()

    # Exclude the target column itself from potential leakage
    potential_leakage.remove(target_column)

    if len(potential_leakage) > 0:
        warnings.warn(
            f"Potential target leakage detected. Consider dropping the feature(s): {potential_leakage}",
            stacklevel=2,
        )
    else:
        print("No leakage has been detected")

    return potential_leakage


def detect_categorical_leakage(
    data: pd.DataFrame, target_column: Union[str, float, int], threshold: float = 0.9
) -> List[Union[str, float, int, None]]:
    """
    Detect data leakage by calculating Theil's U for categorical variables with respect to the target.
    The target column must be part of the provided DataFrame.

    :param data: The DataFrame containing the data.
    :param target_column: The name of the target column.
    :param threshold: The threshold for Theil's U. Columns with U greater than or equal to this threshold
      will be considered potential data leakage.
    :returns: A list of column names with Theil's U greater than or equal to the threshold.
    """
    if target_column not in data.columns:
        raise ValueError(
            f"The target column '{target_column}' is not found in the DataFrame."
        )
    else:
        data[target_column] = data[target_column].astype(str)

    leakage_columns = []
    for column in data.columns:
        if column == target_column:
            continue
        u = theil_u(data[column], data[target_column])
        if u >= threshold:
            leakage_columns.append(column)

    if len(leakage_columns) > 0:
        warnings.warn(
            f"Potential target leakage or constant categorical columns detected. Consider dropping the feature(s): {leakage_columns}",
            stacklevel=2,
        )
    else:
        print("No leakage has been detected")

    return leakage_columns



from typing import Any, List, Literal, Optional, Tuple, Union

import pandas as pd
from sklearn.model_selection import StratifiedKFold

from bluecast.blueprints.cast import BlueCast
from bluecast.config.training_config import (
    TrainingConfig,
    XgboostFinalParamConfig,
    XgboostTuneParamsConfig,
)
from bluecast.experimentation.tracking import ExperimentTracker
from bluecast.general_utils.general_utils import logger
from bluecast.ml_modelling.xgboost import XgboostModel
from bluecast.preprocessing.custom import CustomPreprocessing
from bluecast.preprocessing.feature_selection import RFECVSelector


class BlueCastCV:
    """Wrapper to train and predict multiple blueCast intstances.

    Check the BlueCast class documentation for additional parameter details.
    A custom splitter can be provided.
    """

    def __init__(
        self,
        class_problem: Literal["binary", "multiclass"] = "binary",
        stratifier: Optional[Any] = None,
        conf_training: Optional[TrainingConfig] = None,
        conf_xgboost: Optional[XgboostTuneParamsConfig] = None,
        conf_params_xgboost: Optional[XgboostFinalParamConfig] = None,
        experiment_tracker: Optional[ExperimentTracker] = None,
        custom_in_fold_preprocessor: Optional[CustomPreprocessing] = None,
        custom_last_mile_computation: Optional[CustomPreprocessing] = None,
        custom_preprocessor: Optional[CustomPreprocessing] = None,
        custom_feature_selector: Optional[
            Union[RFECVSelector, CustomPreprocessing]
        ] = None,
        ml_model: Optional[Union[XgboostModel, Any]] = None,
    ):
        self.class_problem = class_problem
        self.conf_xgboost = conf_xgboost
        self.conf_training = conf_training
        self.conf_params_xgboost = conf_params_xgboost
        self.custom_in_fold_preprocessor = custom_in_fold_preprocessor
        self.custom_preprocessor = custom_preprocessor
        self.custom_feature_selector = custom_feature_selector
        self.custom_last_mile_computation = custom_last_mile_computation
        self.bluecast_models: List[BlueCast] = []
        self.stratifier = stratifier
        self.ml_model = ml_model

        if experiment_tracker:
            self.experiment_tracker = experiment_tracker
        else:
            self.experiment_tracker = ExperimentTracker()

    def prepare_data(
        self, df: pd.DataFrame, target: str
    ) -> Tuple[pd.DataFrame, pd.Series]:
        df = df.reset_index(drop=True)
        y = df[target]
        X = df.drop(target, axis=1)
        return X, y

    def fit(self, df: pd.DataFrame, target_col: str) -> None:
        """Fit multiple BlueCast instances on different data splits.

        Input df is expected the target column."""
        X, y = self.prepare_data(df, target_col)

        if not self.conf_training:
            self.conf_training = TrainingConfig()

        if not self.stratifier:
            self.stratifier = StratifiedKFold(
                n_splits=5,
                shuffle=True,
                random_state=self.conf_training.global_random_state,
            )

        for fn, (trn_idx, val_idx) in enumerate(self.stratifier.split(X, y)):
            X_train, X_val = X.iloc[trn_idx], X.iloc[val_idx]
            y_train, y_val = y.iloc[trn_idx], y.iloc[val_idx]
            x_train = pd.concat([X_train, X_val], ignore_index=True)
            y_train = pd.concat([y_train, y_val], ignore_index=True)

            X_train = x_train.reset_index(drop=True)
            y_train = y_train.reset_index(drop=True)
            X_train[target_col] = y_train.values

            self.conf_training.global_random_state += (
                self.conf_training.increase_random_state_in_bluecast_cv_by
            )
            logger(
                f"Start fitting model number {fn} with random seed {self.conf_training.global_random_state}"
            )

            automl = BlueCast(
                class_problem=self.class_problem,
                target_column=target_col,
                conf_training=self.conf_training,
                conf_xgboost=self.conf_xgboost,
                conf_params_xgboost=self.conf_params_xgboost,
                experiment_tracker=self.experiment_tracker,
                custom_in_fold_preprocessor=self.custom_in_fold_preprocessor,
                custom_preprocessor=self.custom_preprocessor,
                custom_feature_selector=self.custom_feature_selector,
                custom_last_mile_computation=self.custom_last_mile_computation,
                ml_model=self.ml_model,
            )
            automl.fit(X_train, target_col=target_col)
            self.bluecast_models.append(automl)

            # overwrite experiment tracker to pass it into next iteration
            self.experiment_tracker = automl.experiment_tracker

    def fit_eval(self, df: pd.DataFrame, target_col: str) -> None:
        """Fit multiple BlueCast instances on different data splits.

        Input df is expected the target column. Evaluation is executed on out-of-fold dataset
        in each split."""
        X, y = self.prepare_data(df, target_col)

        if not self.conf_training:
            self.conf_training = TrainingConfig()

        if not self.stratifier:
            self.stratifier = StratifiedKFold(
                n_splits=5,
                shuffle=True,
                random_state=self.conf_training.global_random_state,
            )

        for fn, (trn_idx, val_idx) in enumerate(self.stratifier.split(X, y)):
            X_train, X_val = X.iloc[trn_idx], X.iloc[val_idx]
            y_train, y_val = y.iloc[trn_idx], y.iloc[val_idx]

            X_train.loc[:, target_col] = y_train

            self.conf_training.global_random_state += (
                self.conf_training.increase_random_state_in_bluecast_cv_by
            )
            logger(
                f"Start fitting model number {fn} with random seed {self.conf_training.global_random_state}"
            )

            automl = BlueCast(
                class_problem=self.class_problem,
                target_column=target_col,
                conf_training=self.conf_training,
                conf_xgboost=self.conf_xgboost,
                conf_params_xgboost=self.conf_params_xgboost,
                experiment_tracker=self.experiment_tracker,
                custom_in_fold_preprocessor=self.custom_in_fold_preprocessor,
                custom_preprocessor=self.custom_preprocessor,
                custom_feature_selector=self.custom_feature_selector,
                custom_last_mile_computation=self.custom_last_mile_computation,
                ml_model=self.ml_model,
            )
            automl.fit_eval(X_train, X_val, y_val, target_col=target_col)
            self.bluecast_models.append(automl)

            # overwrite experiment tracker to pass it into next iteration
            self.experiment_tracker = automl.experiment_tracker

    def predict(
        self, df: pd.DataFrame, return_sub_models_preds: bool = False
    ) -> Tuple[Union[pd.DataFrame, pd.Series], Union[pd.DataFrame, pd.Series]]:
        """Predict on unseen data using multiple trained BlueCast instances"""
        or_cols = df.columns
        prob_cols: list[str] = []
        class_cols: list[str] = []
        for fn, pipeline in enumerate(self.bluecast_models):
            y_probs, y_classes = pipeline.predict(df.loc[:, or_cols])
            if self.class_problem == "multiclass":
                proba_cols = [
                    f"class_{col}_proba_model_{fn}" for col in range(y_probs.shape[1])
                ]
                df[proba_cols] = y_probs
                df[f"classes_{fn}"] = y_classes
                for col in proba_cols:
                    prob_cols.append(col)
                class_cols.append(f"classes_{fn}")

            else:
                df[f"proba_{fn}"] = y_probs
                df[f"classes_{fn}"] = y_classes
                prob_cols.append(f"proba_{fn}")
                class_cols.append(f"classes_{fn}")

        if self.class_problem == "multiclass":
            if return_sub_models_preds:
                return df.loc[:, prob_cols], df.loc[:, class_cols]
            else:
                classes = df.loc[:, class_cols].mode(axis=1)[0].astype(int)

                if self.bluecast_models[0].feat_type_detector:
                    if (
                        self.bluecast_models[0].target_label_encoder
                        and self.bluecast_models[0].feat_type_detector
                    ):
                        classes = self.bluecast_models[
                            0
                        ].target_label_encoder.label_encoder_reverse_transform(classes)

                return (
                    df.loc[:, prob_cols].mean(axis=1),
                    classes,
                )
        else:
            if return_sub_models_preds:
                return df.loc[:, prob_cols], df.loc[:, class_cols]
            else:
                return (
                    df.loc[:, prob_cols].mean(axis=1),
                    df.loc[:, prob_cols].mean(axis=1) > 0.5,
                )



"""Define training and common configuration parameters.

Pydantic dataclasses are used to define the configuration parameters. This allows for type checking and validation of
the configuration parameters. The configuration parameters are used in the training pipeline and in the evaluation
pipeline. Pydantic dataclasses are used to allow users a pythonic way to define the configuration parameters.
Default configurations can be loaded, adjusted and passed into the blueprints.
"""
from typing import Dict, Optional

from pydantic import BaseModel
from pydantic.dataclasses import dataclass


class Config:
    arbitrary_types_allowed = True


class TrainingConfig(BaseModel):
    """Define general training parameters.

    :param global_random_state: Global random state to use for reproducibility.
    :param increase_random_state_in_bluecast_cv_by: In BlueCastCV multiple models are trained. Define by how much the
        random state changes with each additional model.
    :param shuffle_during_training: Whether to shuffle the data during training when hypertuning_cv_folds > 1.
    :param hyperparameter_tuning_rounds: Number of hyperparameter tuning rounds. Not used when custom ML model is passed.
    :param hyperparameter_tuning_max_runtime_secs: Maximum runtime in seconds for hyperparameter tuning. Not used when
        custom ML model is passed.
    :param hypertuning_cv_folds: Number of cross-validation folds to use for hyperparameter tuning. Not used when
        custom ML model is passed.
    :param precise_cv_tuning: If enabled will switch from using Xgboost's own cv method to a custom cross validation
        routine. This is needed when the in-fold preprocessing is necessary that would cause overfitting with usual cv.
        This has a much longer runtime as Optuna's pruning call is missing and all trials will run until the end.
    :param early_stopping_rounds: Number of early stopping rounds. Not used when custom ML model is passed. Also
        not used when hypertuning_cv_folds is greater than 1.
    :param autotune_model: Whether to autotune the model. Not used when custom ML model is passed.
    :param enable_feature_selection: Whether to enable recursive feature selection.
    :param calculate_shap_values: Whether to calculate shap values. Also used when custom ML model is passed. Not
        compatible with all ML models. See the SHAP documentation for more details.
    :param train_size: Train size to use for train-test split.
    :param train_split_stratify: Whether to stratify the train-test split. Not used when custom ML model is passed.
    :param use_full_data_for_final_model: Whether to use the full data for the final model. This might cause overfitting.
        Not used when custom ML model is passed.
    :param min_features_to_select: Minimum number of features to select. Only used when enable_feature_selection is
        True.
    :param cat_encoding_via_ml_algorithm: Whether to use an ML algorithm for categorical encoding. If True, the
        categorical encoding is done via a ML algorithm. If False, the categorical encoding is done via a  target
        encoding in the preprocessing steps. See the ReadMe for more details.
    :param show_detailed_tuning_logs: Whether to show detailed tuning logs. Not used when custom ML model is passed.
    :param enable_grid_search_fine_tuning: After hyperparameter tuning run Gridsearch tuning on a fine-grained grid
        based on the previous hyperparameter tuning. Only possible when autotune_model is True.
    :param gridsearch_nb_parameters_per_grid: Decides how many steps the grid shall have per parameter.
    :param gridsearch_tuning_max_runtime_secs: Sets the maximum time in seconds the tuning shall run. This will finish
        the latest trial nd will exceed this limit though.
    :param experiment_name: Name of the experiment. Will be logged inside the ExperimentTracker.
    """

    global_random_state: int = 10
    increase_random_state_in_bluecast_cv_by: int = 33
    shuffle_during_training: bool = True
    hyperparameter_tuning_rounds: int = 200
    hyperparameter_tuning_max_runtime_secs: int = 3600
    hypertuning_cv_folds: int = 1
    precise_cv_tuning: bool = False
    early_stopping_rounds: int = 10
    autotune_model: bool = True
    enable_feature_selection: bool = False
    calculate_shap_values: bool = True
    train_size: float = 0.8
    train_split_stratify: bool = True
    use_full_data_for_final_model: bool = False
    min_features_to_select: int = 5
    cat_encoding_via_ml_algorithm: bool = False
    show_detailed_tuning_logs: bool = False
    optuna_sampler_n_startup_trials: int = 10
    enable_grid_search_fine_tuning: bool = False
    gridsearch_tuning_max_runtime_secs: int = 3600
    gridsearch_nb_parameters_per_grid: int = 5
    experiment_name: str = "new experiment"


class XgboostTuneParamsConfig(BaseModel):
    """Define hyperparameter tuning search space."""

    max_depth_min: int = 2
    max_depth_max: int = 6
    alpha_min: float = 0.0
    alpha_max: float = 10.0
    lambda_min: float = 0.0
    lambda_max: float = 10.0
    gamma_min: float = 0.0
    gamma_max: float = 10.0
    subsample_min: float = 0.0
    subsample_max: float = 10.0
    max_leaves_min: int = 0
    max_leaves_max: int = 0
    sub_sample_min: float = 0.3
    sub_sample_max: float = 1.0
    col_sample_by_tree_min: float = 0.3
    col_sample_by_tree_max: float = 1.0
    col_sample_by_level_min: float = 0.3
    col_sample_by_level_max: float = 1.0
    eta_min: float = 0.001
    eta_max: float = 0.3
    steps_min: int = 2
    steps_max: int = 1000
    model_verbosity: int = 0
    model_verbosity_during_final_training: int = 0
    model_objective: str = "multi:softprob"
    model_eval_metric: str = "mlogloss"
    booster: str = "gbtree"


@dataclass
class XgboostFinalParamConfig:
    """Define final hyper parameters."""

    params = {
        "objective": "multi:softprob",  # OR  'binary:logistic' #the loss function being used
        "eval_metric": "mlogloss",
        "tree_method": "exact",  # use GPU for training
        "num_class": 2,
        "max_depth": 3,  # maximum depth of the decision trees being trained
        "alpha": 0.1,
        "lambda": 0.1,
        "gamma": 0.0,
        "subsample": 1.0,
        "max_leaves": 16,
        "colsample_bytree": 0.8,
        "colsample_bylevel": 0.8,
        "eta": 0.1,
        "steps": 1000,
        "num_parallel_tree": 1,
    }
    sample_weight: Optional[Dict[str, float]] = None
    classification_threshold: float = 0.5



import random
from typing import Tuple

import pandas as pd
import pytest

from bluecast.eda.analyse import (
    bi_variate_plots,
    check_unique_values,
    correlation_heatmap,
    correlation_to_target,
    mutual_info_to_target,
    plot_null_percentage,
    plot_pca,
    plot_theil_u_heatmap,
    plot_tsne,
    univariate_plots,
)
from bluecast.tests.make_data.create_data import create_synthetic_dataframe


@pytest.fixture
def synthetic_train_test_data() -> Tuple[pd.DataFrame, pd.DataFrame]:
    df_train = create_synthetic_dataframe(2000, random_state=20)
    df_val = create_synthetic_dataframe(2000, random_state=200)
    return df_train, df_val


@pytest.fixture()
def synthetic_categorical_data() -> pd.DataFrame:
    data = {
        "Category1": [random.choice(["A", "B", "C"]) for _ in range(100)],
        "Category2": [random.choice(["X", "Y", "Z"]) for _ in range(100)],
        "Category3": [random.choice(["Red", "Green", "Blue"]) for _ in range(100)],
        "Category4": [random.choice(["High", "Medium", "Low"]) for _ in range(100)],
    }

    # Create the DataFrame
    df = pd.DataFrame(data)
    return df


@pytest.fixture
def create_data_with_nulls() -> pd.DataFrame:
    data = {
        "Column1": [1, 2, 3, None, 5],
        "Column2": [None, 2, 3, 4, 5],
        "Column3": [1, 2, 3, 4, 5],
        "Column4": [None, None, None, None, None],
    }

    df = pd.DataFrame(data)
    return df


@pytest.fixture
def create_data_with_many_uniques() -> pd.DataFrame:
    df = pd.DataFrame(
        {
            "col1": [1, 2, 3, 4],
            "col2": [1, 2, 3, 3],
            "col3": [1, 2, 2, 2],
            "col4": [1, 1, 1, 1],
        }
    )
    return df


def test_univariate_plots(synthetic_train_test_data):
    univariate_plots(
        synthetic_train_test_data[0].loc[
            :, ["numerical_feature_1", "numerical_feature_2", "numerical_feature_3"]
        ],
        "target",
    )
    assert True


def test_bi_variate_plots(synthetic_train_test_data):
    bi_variate_plots(synthetic_train_test_data[0], "target")
    assert True


def test_correlation_heatmap(synthetic_train_test_data):
    correlation_heatmap(
        synthetic_train_test_data[0].drop(
            ["categorical_feature_1", "categorical_feature_2"], axis=1
        )
    )
    assert True


def test_correlation_to_target(synthetic_train_test_data):
    correlation_to_target(
        synthetic_train_test_data[0].drop(
            ["categorical_feature_1", "categorical_feature_2"], axis=1
        ),
        "target",
    )
    assert True


def test_mutual_info_to_target(synthetic_train_test_data):
    mutual_info_to_target(
        synthetic_train_test_data[0].drop(
            ["categorical_feature_1", "categorical_feature_2", "datetime_feature"],
            axis=1,
        ),
        "target",
    )


def test_pca_plot(synthetic_train_test_data):
    plot_pca(
        synthetic_train_test_data[0].loc[
            :,
            [
                "numerical_feature_1",
                "numerical_feature_2",
                "numerical_feature_3",
                "target",
            ],
        ],
        "target",
    )
    assert True


def test_plot_tsne(synthetic_train_test_data):
    plot_tsne(
        synthetic_train_test_data[0].loc[
            :,
            [
                "numerical_feature_1",
                "numerical_feature_2",
                "numerical_feature_3",
                "target",
            ],
        ],
        "target",
        perplexity=30,
        random_state=0,
    )
    assert True


def test_plot_theil_u_heatmap(synthetic_categorical_data):
    columns_of_interest = synthetic_categorical_data.columns.to_list()
    theil_matrix = plot_theil_u_heatmap(synthetic_categorical_data, columns_of_interest)
    assert True
    assert theil_matrix[0, 0] == 1.0


def test_plot_null_percentage(create_data_with_nulls):
    plot_null_percentage(create_data_with_nulls)
    assert True


def test_check_unique_values(create_data_with_many_uniques):
    # Test with threshold of 0.9
    assert check_unique_values(
        create_data_with_many_uniques, ["col1", "col2", "col3", "col4"], 0.9
    ) == ["col1"]

    # Test with threshold of 0.8
    assert check_unique_values(
        create_data_with_many_uniques, ["col1", "col2", "col3"], 0.70
    ) == ["col1", "col2"]

    # Test with threshold of 0.5
    assert check_unique_values(
        create_data_with_many_uniques, ["col1", "col2", "col3"], 0.5
    ) == ["col1", "col2", "col3"]



from unittest import mock

import numpy as np
import pandas as pd
import pytest
from sklearn.linear_model import LogisticRegression

from bluecast.blueprints.cast import BlueCast
from bluecast.config.training_config import TrainingConfig, XgboostTuneParamsConfig
from bluecast.evaluation.eval_metrics import matthews_corrcoef
from bluecast.evaluation.shap_values import shap_explanations
from bluecast.tests.make_data.create_data import create_synthetic_dataframe


@pytest.fixture
def mock_model():
    return mock.Mock()


@pytest.fixture
def test_data():
    return pd.DataFrame({"feature1": [1, 2, 3], "feature2": [4, 5, 6]})


def test_shap_explanations():
    """Test that tests the BlueCast class"""
    df_train = create_synthetic_dataframe(200, random_state=20)
    df_val = create_synthetic_dataframe(100, random_state=21)
    xgboost_param_config = XgboostTuneParamsConfig()
    xgboost_param_config.steps_max = 100
    train_config = TrainingConfig()
    train_config.hyperparameter_tuning_rounds = 10
    train_config.enable_feature_selection = False
    train_config.hypertuning_cv_folds = 2

    automl = BlueCast(
        class_problem="binary",
        target_column="target",
        conf_training=train_config,
        conf_xgboost=xgboost_param_config,
    )
    eval_dict = automl.fit_eval(
        df_train, df_val.drop("target", axis=1), df_val["target"], target_col="target"
    )
    print(eval_dict)
    assert isinstance(eval_dict, dict)

    targets = df_train.pop("target")

    # test that non-tree based models are supported as well
    model = LogisticRegression()
    model.fit(
        df_train.loc[
            :, ["numerical_feature_1", "numerical_feature_2", "numerical_feature_3"]
        ],
        targets,
    )
    model_shap_values = shap_explanations(
        model,
        df_val.loc[
            :, ["numerical_feature_1", "numerical_feature_2", "numerical_feature_3"]
        ],
    )
    assert model_shap_values.any()


def test_eval_classifier_except():
    y_true = np.array([0, 0, 0, 0])
    y_classes = np.array([1, 1, 1, 1])
    result = matthews_corrcoef(y_true, y_classes)
    assert result == 0



import math
from collections import Counter
from typing import List, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.stats as ss
import seaborn as sns
from sklearn.decomposition import PCA
from sklearn.feature_selection import mutual_info_classif
from sklearn.manifold import TSNE


def univariate_plots(df: pd.DataFrame, target: str) -> None:
    """
    Plots univariate plots for all the columns in the dataframe.
    The target column must be part of the provided DataFrame.

    Expects numeric columns only.
    """
    for col in df.columns:
        # Check if the col is the target column (EC1 or EC2)
        if col == target:
            continue  # Skip target columns in univariate analysis

        plt.figure(figsize=(8, 4))

        # Histogram
        plt.subplot(1, 2, 1)
        sns.histplot(data=df, x=col, kde=True)
        plt.xlabel(col)
        plt.ylabel("Frequency")
        plt.title("Histogram")

        # Box plot
        plt.subplot(1, 2, 2)
        sns.boxplot(data=df, y=col)
        plt.ylabel(col)
        plt.title("Box Plot")

        # Adjust spacing between subplots
        plt.tight_layout()

        # Show the plots
        plt.show()


def bi_variate_plots(df: pd.DataFrame, target: str, num_cols_grid: int = 4) -> None:
    """
    Plots bivariate plots for all column combinations in the dataframe.
    The target column must be part of the provided DataFrame.
    Param num_cols_grid specifies how many columns the grid shall have.

    Expects numeric columns only.
    """
    # Get the list of column names except for the target column
    variables = [col for col in df.columns if col != target]

    # Define the grid layout based on the number of variables
    num_variables = len(variables)
    num_cols = num_cols_grid  # Number of columns in the grid
    num_rows = (
        num_variables + num_cols - 1
    ) // num_cols  # Calculate the number of rows needed

    # Set the size of the figure
    fig, axes = plt.subplots(num_rows, num_cols, figsize=(12, 4 * num_rows))

    # Generate violin plots for each variable with respect to EC1
    for i, variable in enumerate(variables):
        row = i // num_cols
        col = i % num_cols
        ax = axes[row][col]

        sns.violinplot(data=df, x=target, y=variable, ax=ax)
        ax.set_xlabel(target)
        ax.set_ylabel(variable)
        ax.set_title(f"Violin Plot: {variable} vs {target}")

    # Remove any empty subplots
    if num_variables < num_rows * num_cols:
        for i in range(num_variables, num_rows * num_cols):
            fig.delaxes(axes.flatten()[i])

    # Adjust the spacing between subplots
    plt.tight_layout()

    # Show the plot
    plt.show()


def correlation_heatmap(df: pd.DataFrame) -> None:
    """
    Plots half of the heatmap showing correlations of all features.

    Expects numeric columns only.
    """
    # Calculate the correlation matrix
    corr = df.corr()

    # Generate a mask for the upper triangle
    mask = np.triu(np.ones_like(corr, dtype=bool))

    # Generate a custom diverging colormap
    cmap = sns.diverging_palette(230, 20, as_cmap=True)

    # Draw the heatmap with the mask and correct aspect ratio
    sns.heatmap(
        corr,
        mask=mask,
        cmap=cmap,
        vmax=0.3,
        center=0,
        square=True,
        linewidths=0.5,
        cbar_kws={"shrink": 0.5},
    )

    plt.show()


def correlation_to_target(df: pd.DataFrame, target: str) -> None:
    """
    Plots correlations for all the columns in the dataframe in relation to the target column.
    The target column must be part of the provided DataFrame.

    Expects numeric columns only.
    """
    # Calculate the correlation matrix
    corr = df.corr()

    # Get correlations without 'EC1' and 'EC2'
    corrs = corr[target].drop([target])

    # Sort correlation values in descending order
    corrs_sorted = corrs.sort_values(ascending=False)

    # Create a heatmap of the correlations with EC1
    sns.set(font_scale=0.8)
    sns.set_style("white")
    sns.set_palette("PuBuGn_d")
    sns.heatmap(corrs_sorted.to_frame(), cmap="coolwarm", annot=True, fmt=".2f")
    plt.title("Correlation with EC1")
    plt.show()


def plot_pca(df: pd.DataFrame, target: str) -> None:
    """
    Plots PCA for the dataframe. The target column must be part of the provided DataFrame.

    Expects numeric columns only.
    """
    pca = PCA(n_components=2)

    pca_df = pd.DataFrame(pca.fit_transform(df.drop([target], axis=1)))
    pca_df[target] = df[target].values
    fig, ax = plt.subplots(ncols=1, figsize=(10, 5))
    explained_variance = round(sum(pca.explained_variance_ratio_), 2)

    # Define a custom color palette with distinct colors for the target variable
    target_palette = sns.color_palette("hls", len(df[target].unique()))

    for _i, col in enumerate([target]):
        sns.scatterplot(data=pca_df, x=0, y=1, hue=col, ax=ax, palette=target_palette)
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_xlabel("Component 1")
        ax.set_ylabel("Component 2")

    fig.suptitle(f"PCA \n explained variance :{explained_variance}", y=1.1)
    plt.show()


def plot_tsne(df: pd.DataFrame, target: str, perplexity=50, random_state=42) -> None:
    """
    Plots t-SNE for the dataframe. The target column must be part of the provided DataFrame.

    Expects numeric columns only.
    """
    tsne = TSNE(n_components=2, perplexity=perplexity, random_state=random_state)

    tsne_df = pd.DataFrame(tsne.fit_transform(df.drop([target], axis=1)))
    tsne_df[target] = df[target].values
    fig, ax = plt.subplots(ncols=1, figsize=(10, 5))

    # Define a custom color palette with distinct colors for the target variable
    target_palette = sns.color_palette("hls", len(df[target].unique()))

    for _i, col in enumerate([target]):
        sns.scatterplot(data=tsne_df, x=0, y=1, hue=col, ax=ax, palette=target_palette)
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_xlabel("Component 1")
        ax.set_ylabel("Component 2")

    fig.suptitle("t-SNE", y=1.1)
    plt.show()


def conditional_entropy(x, y):
    # entropy of x given y
    y_counter = Counter(y)
    xy_counter = Counter(list(zip(x, y)))
    total_occurrences = sum(y_counter.values())
    entropy = 0
    for xy in xy_counter.keys():
        p_xy = xy_counter[xy] / total_occurrences
        p_y = y_counter[xy[1]] / total_occurrences
        entropy += p_xy * math.log(p_y / p_xy)
    return entropy


def theil_u(x, y):
    s_xy = conditional_entropy(x, y)
    x_counter = Counter(x)
    total_occurrences = sum(x_counter.values())
    p_x = list(map(lambda n: n / total_occurrences, x_counter.values()))
    s_x = ss.entropy(p_x)
    if s_x == 0:
        return 1
    else:
        return (s_x - s_xy) / s_x


def plot_theil_u_heatmap(data: pd.DataFrame, columns: List[Union[str, int, float]]):
    """Plot a heatmap for categorical data using Theil's U."""
    theil_matrix = np.zeros((len(columns), len(columns)))

    for i in range(len(columns)):
        for j in range(len(columns)):
            theil_matrix[i, j] = theil_u(data[columns[i]], data[columns[j]])

    plt.figure(figsize=(len(columns), len(columns)))
    sns.heatmap(
        theil_matrix,
        annot=True,
        xticklabels=columns,
        yticklabels=columns,
        cmap="coolwarm",
    )
    plt.title("Theil's U Heatmap")

    plt.show()
    return theil_matrix


def plot_null_percentage(dataframe: pd.DataFrame) -> None:
    # Calculate the percentage of null values for each column
    null_percentage = (dataframe.isnull().mean() * 100).round(2)

    # Create a bar plot to visualize the null percentages
    plt.figure(figsize=(12, 6))
    bars = plt.bar(null_percentage.index, null_percentage.values)
    plt.title("Percentage of Null Values in Each Column")
    plt.xlabel("Columns")
    plt.ylabel("Percentage of Null Values")
    plt.xticks(rotation=90)
    plt.tight_layout()

    # Add annotations to the bars
    for bar, percentage in zip(bars, null_percentage.values):
        plt.text(
            bar.get_x() + bar.get_width() / 2 - 0.15,  # Adjust x-position for centering
            bar.get_height() + 1,  # Adjust y-position for vertical alignment
            f"{percentage}%",  # Display the percentage
            ha="center",  # Horizontal alignment
            va="bottom",  # Vertical alignment
        )

    # Show the plot
    plt.show()


def check_unique_values(
    df: pd.DataFrame, columns: List[Union[str, int, float]], threshold: float
) -> List[Union[str, int, float]]:
    """
    Check if the columns have an amount of unique values that is almost the number of total rows (being above the defined threshold)

    :param df: The pandas DataFrame to check
    :param columns: A list of column names to check
    :param threshold: The threshold to check against
    :returns: A list of column names that have a high amount of unique values
    """
    total_rows = len(df.index)
    lots_uniques = []
    for column in columns:
        unique_values = len(df[column].unique())
        if unique_values / total_rows >= threshold:
            lots_uniques.append(column)
    return lots_uniques


def mutual_info_to_target(df: pd.DataFrame, target: str, **mut_params) -> None:
    """
    Plots mutual information scores for all the categorical columns in the DataFrame in relation to the target column.
    The target column must be part of the provided DataFrame.
    :param df: DataFrame containing all columns including target column. Features are expected to be numerical.
    :param target: String indicating which column is the target column.
    :param mut_params: Dictionary passing additional arguments into sklearn's mutual_info_classif function.

    To be used for classification only.
    """
    # Calculate the mutual information scores
    mi_scores = mutual_info_classif(
        X=df.drop(columns=[target]), y=df[target], **mut_params
    )

    # Sort features by MI score descending
    sorted_features = df.drop(columns=[target]).columns[np.argsort(-mi_scores)]

    # Sort MI scores in descending order
    mi_scores_sorted = mi_scores[np.argsort(-mi_scores)]

    # Create a bar chart of the mutual information scores
    fig, ax = plt.subplots(figsize=(10, 6))
    sns.barplot(x=mi_scores_sorted, y=sorted_features, order=sorted_features, ax=ax)
    ax.set_title("Mutual Information Scores with Target")
    ax.set_xlabel("Mutual Information Score")
    ax.set_ylabel("Features")
    for i, v in enumerate(mi_scores_sorted):
        ax.text(v + 0.01, i, str(round(v, 2)), color="blue", fontweight="bold")
    plt.show()



import numpy as np
import pandas as pd

from bluecast.preprocessing.nulls_and_infs import fill_infinite_values, fill_nulls


def test_fill_nulls():
    # Create a test DataFrame with null values
    df = pd.DataFrame({"A": [1, None, 3, 4], "B": [None, 6, 7, None]})

    # Test case 1: Fill null values with 0
    expected_output = pd.DataFrame({"A": [1, 0, 3, 4], "B": [0, 6, 7, 0]})
    assert (
        fill_nulls(df, fill_with=0).astype(float).equals(expected_output.astype(float))
    )

    # Test case 2: Fill null values with NaN
    df_alt = pd.DataFrame(
        {"A": [1, float("nan"), 3, 4], "B": [float("nan"), 6, 7, float("nan")]}
    )
    assert fill_nulls(df_alt).astype(float).equals(expected_output.astype(float))


def test_fill_infinite_values():
    # Create a sample DataFrame with infinite values
    df = pd.DataFrame({"A": [1, 2, np.inf, 4], "B": [np.inf, 6, -np.inf, 8]})

    # Test with default fill_with value (0)
    expected_result = pd.DataFrame({"A": [1, 2, 0, 4], "B": [0, 6, 0, 8]})
    assert fill_infinite_values(df.astype(float)).equals(expected_result.astype(float))

    # Test with a different fill_with value (10)
    expected_result = pd.DataFrame({"A": [1, 2, 10, 4], "B": [10, 6, 10, 8]})
    assert fill_infinite_values(df, fill_with=10).equals(expected_result.astype(float))

    # Test with a DataFrame that does not contain infinite values
    df = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
    expected_result = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})
    assert fill_infinite_values(df).equals(expected_result)



"""
Module for extracting date parts from datetime columns.

Cyclic transformations are not implemented as they aren't a good match for tree models.
"""
from datetime import datetime
from typing import List, Optional, Union

import pandas as pd

from bluecast.general_utils.general_utils import logger


def date_converter(
    df: pd.DataFrame,
    date_columns: Optional[List[Union[str, int, float]]],
    date_parts: List[str] = None,
) -> pd.DataFrame:
    """
    Takes in a df and loops through datetime columns to and extracts the date parts month, day, dayofweek
    and hour and adds them as additional columns.
    :param date_columns: List of datetime columns.
    :param df: Dataframe to be processed.
    :param date_parts: List of date parts to be extracted.
    :return: Returns modified df.
    """
    logger(f"{datetime.utcnow()}: Start date column conversion.")
    if not date_columns:
        return df

    if not date_parts:
        date_parts = ["month", "day", "dayofweek", "hour"]

    for c in date_columns:
        if "month" in date_parts:
            df[str(c) + "_month"] = df[c].dt.month
        if "day" in date_parts:
            df[str(c) + "_day"] = df[c].dt.day
        if "dayofweek" in date_parts:
            df[str(c) + "_dayofweek"] = df[c].dt.dayofweek
        if "hour" in date_parts:
            df[str(c) + "_hour"] = df[c].dt.hour
        df = df.drop(c, axis=1)
    return df



from typing import Tuple

import numpy as np
import pandas as pd
import pytest
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import StratifiedKFold

from bluecast.blueprints.cast import BlueCast
from bluecast.blueprints.cast_cv import BlueCastCV
from bluecast.config.training_config import TrainingConfig, XgboostTuneParamsConfig
from bluecast.ml_modelling.base_classes import (
    BaseClassMlModel,
    PredictedClasses,
    PredictedProbas,
)
from bluecast.tests.make_data.create_data import create_synthetic_dataframe


@pytest.fixture
def synthetic_train_test_data() -> Tuple[pd.DataFrame, pd.DataFrame]:
    df_train = create_synthetic_dataframe(2000, random_state=20)
    df_val = create_synthetic_dataframe(2000, random_state=200)
    return df_train, df_val


def test_blueprint_cv_xgboost(synthetic_train_test_data):
    """Test that tests the BlueCast cv class"""
    df_train = synthetic_train_test_data[0]
    df_val = synthetic_train_test_data[1]
    xgboost_param_config = XgboostTuneParamsConfig()
    xgboost_param_config.steps_max = 100
    train_config = TrainingConfig()
    train_config.hyperparameter_tuning_rounds = 10

    nb_models = 3

    skf = StratifiedKFold(
        n_splits=nb_models,
        shuffle=True,
        random_state=5,
    )

    automl_cv = BlueCastCV(
        conf_xgboost=xgboost_param_config, conf_training=train_config, stratifier=skf
    )
    automl_cv.fit_eval(
        df_train,
        target_col="target",
    )
    print(automl_cv.experiment_tracker.experiment_id)
    assert (
        len(automl_cv.experiment_tracker.experiment_id)
        <= train_config.hyperparameter_tuning_rounds * nb_models
        + nb_models * 7  # 7 metrics stored in fit_eval
    )
    assert automl_cv.experiment_tracker.experiment_id[-1] < 50
    print("Autotuning successful.")
    y_probs, y_classes = automl_cv.predict(df_val.drop("target", axis=1))
    print("Predicting successful.")
    assert len(y_probs) == len(df_val.index)
    assert len(y_classes) == len(df_val.index)
    y_probs, y_classes = automl_cv.predict(
        df_val.drop("target", axis=1), return_sub_models_preds=True
    )
    assert isinstance(y_probs, pd.DataFrame)
    assert isinstance(y_classes, pd.DataFrame)

    # Assert that the bluecast_models attribute is updated
    assert len(automl_cv.bluecast_models) == nb_models

    automl_cv = BlueCastCV(
        conf_xgboost=xgboost_param_config, conf_training=train_config, stratifier=None
    )
    automl_cv.fit_eval(
        df_train,
        target_col="target",
    )
    assert automl_cv.stratifier
    assert (
        len(automl_cv.experiment_tracker.experiment_id)
        <= train_config.hyperparameter_tuning_rounds * 5
        + 5 * 7  # 7 metrics stored in fit_eval, 5 = default split
    )  # due to Optuna pruning

    # Assert that the bluecast_models attribute is updated
    assert len(automl_cv.bluecast_models) == 5

    # Check if each model in bluecast_models is an instance of BlueCast
    for model in automl_cv.bluecast_models:
        assert isinstance(model, BlueCast)


class CustomLRModel(BaseClassMlModel):
    def __init__(self):
        self.model = None

    def fit(
        self,
        x_train: pd.DataFrame,
        x_test: pd.DataFrame,
        y_train: pd.Series,
        y_test: pd.Series,
    ) -> None:
        self.model = RandomForestClassifier()
        self.model.fit(x_train, y_train)

    def predict(self, df: pd.DataFrame) -> Tuple[PredictedProbas, PredictedClasses]:
        predicted_probas = self.model.predict_proba(df)[:, 1]
        predicted_classes = self.model.predict(df)
        return predicted_probas, predicted_classes


def test_bluecast_cv_fit_eval_with_custom_model():
    custom_model = CustomLRModel()

    # Create an instance of the BlueCast class with the custom model
    bluecast = BlueCast(
        class_problem="binary",
        target_column="target",
        ml_model=custom_model,
    )

    # Create some sample data for testing
    x_train = pd.DataFrame(
        {
            "feature1": [i for i in range(10)],
            "feature2": [i for i in range(10)],
            "feature3": [i for i in range(10)],
            "feature4": [i for i in range(10)],
            "feature5": [i for i in range(10)],
            "feature6": [i for i in range(10)],
        }
    )
    y_train = pd.Series([0, 1, 0, 1, 0, 1, 0, 1, 0, 1])
    x_test = pd.DataFrame(
        {
            "feature1": [i for i in range(10)],
            "feature2": [i for i in range(10)],
            "feature3": [i for i in range(10)],
            "feature4": [i for i in range(10)],
            "feature5": [i for i in range(10)],
            "feature6": [i for i in range(10)],
        }
    )
    y_test = pd.Series([0, 1, 0, 1, 0, 1, 0, 1, 0, 1])

    x_train["target"] = y_train

    # Fit the BlueCast model using the custom model
    bluecast.fit_eval(x_train, x_test, y_test, "target")

    # Predict on the test data using the custom model
    predicted_probas, predicted_classes = bluecast.predict(x_test)

    # Assert the expected results
    assert isinstance(predicted_probas, np.ndarray)
    assert isinstance(predicted_classes, np.ndarray)
    print(bluecast.experiment_tracker.experiment_id)
    assert len(bluecast.experiment_tracker.experiment_id) == 8  # due to custom model



"""
This module contains functions to handle nulls and infinite values.

Only the handling of infinite values is part of the preprocessing pipeline as Xgboost can handle missing values out of
the box.
"""
from datetime import datetime
from typing import Union

import numpy as np
import pandas as pd

from bluecast.general_utils.general_utils import logger


def fill_infinite_values(df: pd.DataFrame, fill_with: Union[int, float] = 0):
    """Replace infinite values with NaN or given value."""
    logger(f"{datetime.utcnow()}: Start filling infinite values.")
    df = df.replace([np.inf, -np.inf], fill_with)
    return df


def fill_nulls(df: pd.DataFrame, fill_with: Union[int, float] = 0):
    """Replace null values with NaN or given value."""
    logger(f"{datetime.utcnow()}: Start filling infinite nulls.")
    df = df.fillna(fill_with)
    return df



import numpy as np
import pandas as pd
from sklearn.datasets import make_classification


def create_synthetic_dataframe(
    num_samples=1000, random_state: int = 20
) -> pd.DataFrame:
    # Generate synthetic data using make_classification
    x, y = make_classification(
        n_samples=num_samples,
        n_features=20,
        n_informative=20,
        n_redundant=0,
        n_clusters_per_class=1,
        random_state=random_state,
    )

    # Create a datetime feature
    start_date = pd.to_datetime("2022-01-01")
    end_date = pd.to_datetime("2022-12-31")
    datetime_feature = pd.date_range(
        start=start_date, end=end_date, periods=num_samples
    )

    # Create categorical features
    categorical_feature_1 = np.random.choice(["A", "B", "C"], size=num_samples)
    categorical_feature_2 = np.random.choice(["X", "Y", "Z"], size=num_samples)

    # Create a DataFrame
    df = pd.DataFrame(
        {
            "categorical_feature_1": categorical_feature_1,
            "categorical_feature_2": categorical_feature_2,
            "numerical_feature_1": x[:, 0],
            "numerical_feature_2": x[:, 1],
            "numerical_feature_3": x[:, 2],
            "datetime_feature": datetime_feature,
            "target": y,
        }
    )

    return df


def create_synthetic_multiclass_dataframe(
    num_samples=1000, random_state: int = 20
) -> pd.DataFrame:
    # Generate synthetic data using make_classification
    x, y = make_classification(
        n_samples=num_samples,
        n_features=20,
        n_informative=20,
        n_redundant=0,
        n_clusters_per_class=1,
        random_state=random_state,
        n_classes=5,
    )

    # Create a datetime feature
    start_date = pd.to_datetime("2022-01-01")
    end_date = pd.to_datetime("2022-12-31")
    datetime_feature = pd.date_range(
        start=start_date, end=end_date, periods=num_samples
    )

    # Create categorical features
    categorical_feature_1 = np.random.choice(["A", "B", "C"], size=num_samples)
    categorical_feature_2 = np.random.choice(["X", "Y", "Z"], size=num_samples)

    # Create a DataFrame
    df = pd.DataFrame(
        {
            "categorical_feature_1": categorical_feature_1,
            "categorical_feature_2": categorical_feature_2,
            "numerical_feature_1": x[:, 0],
            "numerical_feature_2": x[:, 1],
            "numerical_feature_3": x[:, 2],
            "datetime_feature": datetime_feature,
            "target": y,
        }
    )

    return df



import pytest

from bluecast.config.training_config import (
    TrainingConfig,
    XgboostFinalParamConfig,
    XgboostTuneParamsConfig,
)
from bluecast.experimentation.tracking import ExperimentTracker
from bluecast.ml_modelling.xgboost import XgboostModel
from bluecast.tests.make_data.create_data import create_synthetic_dataframe


# Create a fixture for the XGBoost model
@pytest.fixture
def xgboost_model():
    return XgboostModel(class_problem="binary")


# Test case to check if fine-tuning runs without errors
def test_fine_tune_runs_without_errors(xgboost_model):
    xgboost_params = XgboostFinalParamConfig()

    xgboost_model.conf_params_xgboost = xgboost_params
    xgboost_model.conf_training = TrainingConfig()
    xgboost_model.conf_xgboost = XgboostTuneParamsConfig()
    print(xgboost_model.conf_params_xgboost.params)
    xgboost_model.experiment_tracker = ExperimentTracker()
    xgboost_model.conf_training.autotune_model = False

    df_train, df_val = create_synthetic_dataframe(
        2000, random_state=20
    ), create_synthetic_dataframe(2000, random_state=200)
    df_train = df_train.drop(
        ["categorical_feature_1", "categorical_feature_2", "datetime_feature"], axis=1
    )
    df_val = df_val.drop(
        ["categorical_feature_1", "categorical_feature_2", "datetime_feature"], axis=1
    )

    x_train = df_train.drop("target", axis=1)
    y_train = df_train["target"]
    x_test = df_val.drop("target", axis=1)
    y_test = df_val["target"]

    xgboost_model.fine_tune(x_train, x_test, y_train, y_test)
    assert (
        (xgboost_model.conf_params_xgboost.params["alpha"] != 0.1)
        or (xgboost_model.conf_params_xgboost.params["lambda"] != 0.1)
        or (xgboost_model.conf_params_xgboost.params["eta"] != 0.1)
    )


def test_fine_tune_runs_without_errors_using_cv(xgboost_model):
    xgboost_params = XgboostFinalParamConfig()

    xgboost_model.conf_params_xgboost = xgboost_params
    xgboost_model.conf_training = TrainingConfig()
    xgboost_model.conf_xgboost = XgboostTuneParamsConfig()
    print(xgboost_model.conf_params_xgboost.params)
    xgboost_model.experiment_tracker = ExperimentTracker()
    xgboost_model.conf_training.autotune_model = False
    xgboost_model.conf_training.hypertuning_cv_folds = 5  # enable cross validation

    df_train, df_val = create_synthetic_dataframe(
        2000, random_state=20
    ), create_synthetic_dataframe(2000, random_state=200)
    df_train = df_train.drop(
        ["categorical_feature_1", "categorical_feature_2", "datetime_feature"], axis=1
    )
    df_val = df_val.drop(
        ["categorical_feature_1", "categorical_feature_2", "datetime_feature"], axis=1
    )

    x_train = df_train.drop("target", axis=1)
    y_train = df_train["target"]
    x_test = df_val.drop("target", axis=1)
    y_test = df_val["target"]

    xgboost_model.fine_tune(x_train, x_test, y_train, y_test)
    assert (
        (xgboost_model.conf_params_xgboost.params["alpha"] != 0.1)
        or (xgboost_model.conf_params_xgboost.params["lambda"] != 0.1)
        or (xgboost_model.conf_params_xgboost.params["eta"] != 0.1)
    )



"""Base classes for all ML models."""
from abc import ABC, abstractmethod
from typing import Any, Optional, Tuple, TypeVar

import numpy as np
import pandas as pd

PredictedProbas = TypeVar("PredictedProbas", np.ndarray, pd.Series)
PredictedClasses = TypeVar("PredictedClasses", np.ndarray, pd.Series)


class BaseClassMlModel(ABC):
    """Base class for all ML models.

    Enforces the implementation of the fit and predict methods.
    If hyperparameter tuning is required, then the fit method should implement the tuning.
    """

    @abstractmethod
    def fit(
        self,
        x_train: pd.DataFrame,
        x_test: pd.DataFrame,
        y_train: pd.Series,
        y_test: pd.Series,
    ) -> Optional[Any]:
        pass

    @abstractmethod
    def predict(self, df: pd.DataFrame) -> Tuple[PredictedProbas, PredictedClasses]:
        """
        Predict on unseen data.

        :return tuple of predicted probabilities and predicted classes
        """
        pass



import numpy as np
import pandas as pd

from bluecast.preprocessing.feature_types import FeatureTypeDetector


def test_feature_type_detector():
    # Create a sample dataframe for testing
    df = pd.DataFrame(
        {
            "numeric_col1": [1, 2, 3],
            "numeric_col2": [4, 5, 6],
            "boolean_col1": [True, False, True],
            "boolean_col2": [False, True, False],
            "date_col1": ["2021-01-01", "2021-02-01", "2021-03-01"],
            "date_col2": ["2022-01-01", "2022-02-01", "2022-03-01"],
            "categorical_col1": ["cat", "dog", "cat"],
            "categorical_col2": ["dog", "cat", "dog"],
        }
    )

    # Initialize the FeatureTypeDetector object
    detector = FeatureTypeDetector(
        num_columns=["numeric_col1", "numeric_col2"],
        cat_columns=["categorical_col1", "categorical_col2"],
        date_columns=["date_col1", "date_col2"],
    )

    # Test fit_transform_feature_types
    transformed_df = detector.fit_transform_feature_types(df)

    # Numeric columns should remain the same
    assert transformed_df["numeric_col1"].equals(df["numeric_col1"])
    assert transformed_df["numeric_col2"].equals(df["numeric_col2"])

    # Boolean columns should be cast to bool type
    assert transformed_df["boolean_col1"].dtype == np.bool_
    assert transformed_df["boolean_col2"].dtype == np.bool_

    # Date columns should be cast to datetime[ns] type
    assert transformed_df["date_col1"].dtype == np.dtype("datetime64[ns]")
    assert transformed_df["date_col2"].dtype == np.dtype("datetime64[ns]")

    # Categorical columns should remain the same
    assert transformed_df["categorical_col1"].equals(df["categorical_col1"])
    assert transformed_df["categorical_col2"].equals(df["categorical_col2"])

    # Test transform_feature_types
    transformed_df = detector.transform_feature_types(transformed_df)

    # All columns should have the expected data types
    assert transformed_df["numeric_col1"].dtype in ["int64", "float64"]
    assert transformed_df["numeric_col2"].dtype in ["int64", "float64"]
    assert transformed_df["boolean_col1"].dtype == np.bool_
    assert transformed_df["boolean_col2"].dtype == np.bool_
    assert transformed_df["date_col1"].dtype == np.dtype("datetime64[ns]")
    assert transformed_df["date_col2"].dtype == np.dtype("datetime64[ns]")
    assert transformed_df["categorical_col1"].dtype == np.object_
    assert transformed_df["categorical_col2"].dtype == np.object_

    # Ensure the transformed dataframe is the same as the original dataframe
    assert transformed_df.equals(df)



import pandas as pd
import pytest

from bluecast.preprocessing.custom import CustomPreprocessing


@pytest.fixture
def custom_preprocessing():
    class CustomPreprocessingTest(CustomPreprocessing):
        def fit_transform(self, df: pd.DataFrame, target: pd.Series = None):
            return df, target

        def transform(
            self,
            df: pd.DataFrame,
            target: pd.Series = None,
            predicton_mode: bool = False,
        ):
            return df, target

    return (
        CustomPreprocessingTest()
    )  # Replace with your implementation of CustomPreprocessing


def test_fit_transform_returns_tuple(custom_preprocessing):
    df = pd.DataFrame({"col1": [1, 2, 3], "col2": [4, 5, 6]})
    target = pd.Series([0, 1, 0])
    result = custom_preprocessing.fit_transform(df, target)
    assert isinstance(result, tuple)


def test_transform_returns_tuple(custom_preprocessing):
    df = pd.DataFrame({"col1": [1, 2, 3], "col2": [4, 5, 6]})
    target = pd.Series([0, 1, 0])
    result = custom_preprocessing.transform(df, target)
    assert isinstance(result, tuple)


def test_transform_returns_optional_target(custom_preprocessing):
    df = pd.DataFrame({"col1": [1, 2, 3], "col2": [4, 5, 6]})
    result = custom_preprocessing.transform(df)
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert isinstance(result[0], pd.DataFrame)
    assert result[1] is None


def test_transform_returns_optional_target_prediction_mode(custom_preprocessing):
    df = pd.DataFrame({"col1": [1, 2, 3], "col2": [4, 5, 6]})
    result = custom_preprocessing.transform(df, predicton_mode=True)
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert isinstance(result[0], pd.DataFrame)
    assert result[1] is None



"""Module to calculate SHAP values for a trained ML model.

The implementation is flexible and can be used for almost any ML model. The implementation is based on the SHAP library.
"""
import numpy as np
import pandas as pd
import shap


def shap_explanations(model, df: pd.DataFrame) -> np.ndarray:
    """
    See explanations under:
    https://medium.com/rapids-ai/gpu-accelerated-shap-values-with-xgboost-1-3-and-rapids-587fad6822
    :param model: Trained ML model
    :param df: Test data to predict on.
    :return: Shap values
    """
    shap.initjs()
    try:
        tree_explainer = shap.TreeExplainer(model)
        model_shap_values = tree_explainer.shap_values(df)
        shap.summary_plot(model_shap_values, df, plot_type="bar", show=True)
    except (AssertionError, shap.utils._exceptions.InvalidModelError):
        model_shap_explainer = shap.KernelExplainer(model.predict, df)
        model_shap_values = model_shap_explainer.shap_values(df)
        shap.summary_plot(model_shap_values, df, show=True)
    return model_shap_values



"""This module contains the CustomPreprocessing class. This is an entry point for last mile computations before model
training or tuning."""
from abc import ABC, abstractmethod
from typing import Any, Optional, Tuple

import pandas as pd


class CustomPreprocessing(ABC):
    """This class is an entry point for last mile computations before model training or tuning. It is an abstract class
    and must be extended by the user. For fit_transform x_train and y_train are passed. For transform x_test and y_test
    are passed in the BlueCast pipeline. Use prediction_mode = False to skip processing the missing targets.
    :param trained_patterns: Optional. Can we used to save anything from training to be loaded and used in transform.
    If more placeholders are needed, use a dictionary.
    """

    def __init__(self, trained_patterns: Optional[Any] = None):
        self.trained_patterns = trained_patterns

    @abstractmethod
    def fit_transform(
        self, df: pd.DataFrame, target: pd.Series
    ) -> Tuple[pd.DataFrame, pd.Series]:
        return df, target

    @abstractmethod
    def transform(
        self,
        df: pd.DataFrame,
        target: Optional[pd.Series] = None,
        predicton_mode: bool = False,
    ) -> Tuple[pd.DataFrame, Optional[pd.Series]]:
        """Use prediction mode to not process the missing target during inference."""
        return df, target
